/**************************************************************************
***    
***    Copyright (c) 1995-1999 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA, 
***    and by the Regents of the University of California.    
***    All rights reserved.   
***
***    No portion of this package may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

// Created by Andrew Caldwell on Apr 27, 1999

//iterative partitioner calls FM several times, with decreasing tolerance
// if given a problem where there are movable cells if size > 1/2 the 
// original tolerance.

#include "iterTolPart.h"

IterativeTolPart::IterativeTolPart
		(PartitioningProblem& problem, Parameters params)
    : _problem(problem), _params(params), 
      _origMaxCap(problem.getMaxCapacities()),
      _origMinCap(problem.getMinCapacities())
{
    params.fmParams.maxNumPasses = _params.numPassesPerFM;
    const HGraphFixed& hgraph = _problem.getHGraph();

    const vector<vector<double> >& targets = _problem.getCapacities();
    const vector<vector<double> >& maxCapsCon = _problem.getMaxCapacities();
    const vector<vector<double> >& minCapsCon = _problem.getMinCapacities();
    vector<vector<double> >& maxCaps = 
			const_cast<vector<vector<double> >&>(maxCapsCon);
    vector<vector<double> >& minCaps = 
			const_cast<vector<vector<double> >&>(minCapsCon);


    cout<<"Targets:         "<<_problem.getCapacities()<<endl<<endl;
    cout<<"Original MaxCap: "<<_origMaxCap[0][0]<<"  "
			     <<_origMaxCap[1][0]<<endl;
    cout<<"Original MinCap: "<<_origMinCap[0][0]<<"  "
                             <<_origMinCap[1][0]<<endl<<endl;

    abkfatal(_origMaxCap[0][0] == _origMaxCap[1][0] &&
	     _origMinCap[0][0] == _origMinCap[1][0],
	"asymetric max and min values");

    double origTol = _origMaxCap[0][0] - _origMinCap[0][0];
    cout<<"Original Tolerance: "<<origTol<<endl<<endl;


    double largestCell = 0;
    itHGFNodeGlobal n;
    for(n = hgraph.nodesBegin(); n != hgraph.nodesEnd(); n++)
 	largestCell = max(largestCell, (*n)->getWeight()); 

    double newTol = _params.initTolRatio * largestCell;
    cout<<endl;
    cout<<"Largest Cell: "<<largestCell<<endl;
    cout<<"initTolRatio:  "<<_params.initTolRatio<<endl;

    maxCaps[0][0] = targets[0][0] + newTol/2.0;
    minCaps[0][0] = targets[0][0] - newTol/2.0;
    maxCaps[1][0] = targets[1][0] + newTol/2.0;
    minCaps[1][0] = targets[1][0] - newTol/2.0;
 
    cout<<"First MaxCap: "<<_problem.getMaxCapacities()[0][0]<<"  "
			  <<_problem.getMaxCapacities()[1][0]<<endl;    
    cout<<"First MinCap: "<<_problem.getMinCapacities()[0][0]<<"  "
			  <<_problem.getMinCapacities()[1][0]<<endl;    

    newTol = _problem.getMaxCapacities()[0][0] - 
	     _problem.getMinCapacities()[0][0];
    cout<<"New Tolerance: "<<newTol<<endl;

    bool meetsOrigTol = false;
    unsigned iterNum = 0;

    params.fmParams.verb = Verbosity("0 0 0");

    while(!meetsOrigTol)
    {
	cout<<"____Partitioning "<<iterNum<<"_______"<<endl;
	
        FMPartitioner fmpart(_problem, params.fmParams);

     	vector<double> partAreas(2,0.0);
    	const Partitioning& bestSoln = _problem.getBestSoln();

    	for(n = hgraph.nodesBegin(); n != hgraph.nodesEnd(); n++)
    	{
	  unsigned nId = (*n)->getIndex();

	  if(bestSoln[nId].isInPart(0))	
	    partAreas[0] += (*n)->getWeight();
	  else if(bestSoln[nId].isInPart(1))	
	    partAreas[1] += (*n)->getWeight();
	  else
	    abkfatal(0, "node not in p0 or p1");
    	}


	if(partAreas[0] <= _origMaxCap[0][0] &&
	   partAreas[0] >= _origMinCap[0][0] &&
	   partAreas[1] <= _origMaxCap[1][0] &&
	   partAreas[1] >= _origMinCap[1][0]) 
	{
	    meetsOrigTol = true;
	}
	else
	{
	    double largestPart  = max(partAreas[0], partAreas[1]);
	    double smallestPart = min(partAreas[0], partAreas[1]);

  	    maxCaps[0][0] = (largestPart +_origMaxCap[0][0])/2.0; 
  	    maxCaps[1][0] = (largestPart +_origMaxCap[1][0])/2.0; 
  	    minCaps[0][0] = (smallestPart+_origMinCap[0][0])/2.0; 
  	    minCaps[1][0] = (smallestPart+_origMinCap[1][0])/2.0; 

    	    newTol = _problem.getMaxCapacities()[0][0] - 
	     	     _problem.getMinCapacities()[0][0];

	cout<<"      Cut: "
			<<_problem.getCosts()[_problem.getBestSolnNum()]<<endl;
        cout<<"    Areas: "<<partAreas;
    	cout<<"  New Tol: "<<newTol<<endl;
        }
    }
}

ostream& operator<<(ostream& os, const Parameters& params)
{
   os<<" Iterative Tolerance Partitioner Parameters"<<endl;
   os<<" initTolRatio:   "<<params.initTolRatio<<endl;
   os<<" numPassesPerFM: "<<params.numPassesPerFM<<endl;
   os<<endl<<params.fmParams<<endl;
  
   return os;
}
