/**************************************************************************
***
***    Copyright (c) 1995-1998 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

#include "boxPlProb.h"
#include "fxTypeConstr.h"
#include "boxplace.h"

BoxPlacementForCapo::BoxPlacementForCapo
    (   const vector<CapoBlock*>&	allBlocks,
        const HGraphFixed&		hgraphOfNetlist,
	const Placement&		origPl,
	const BBoxPlacer::Parameters&   params)
    :   _allBlocks(allBlocks), _netlistHG(hgraphOfNetlist),
	_boxPl(NULL)
{
	//note:  the first portion of cHG is EXACTLY like the 
	//netlist hypergraph.  That is, the nodes and edges are
	//in the same order.  The constraining nodes and edges have
	//been appended.

    vector<unsigned> nodesToFix;
    nodesToFix.reserve(_netlistHG.getNumTerminals() + _allBlocks.size()*4);

    ConstrainingHGraph cHG(_allBlocks, _netlistHG);
    abkfatal(cHG.getNumTerminals() == _netlistHG.getNumTerminals(),
		"numTerminals should match");
    abkfatal(cHG.getNumNodes() == _netlistHG.getNumNodes() + 
				_allBlocks.size()*4,
		"numNodes in ConstrainingHG is incorrect");

    Placement 	       pl(cHG.getNumNodes());
    for(unsigned n = 0; n < _netlistHG.getNumTerminals(); n++)
    {
	pl[n] = origPl[n];
	nodesToFix.push_back(n);
    }
    for(; n < _netlistHG.getNumNodes(); n++)
	pl[n] = origPl[n];


	//the constraining nodes begin from here. There are four
	//of them for each block.
    unsigned firstConstNode = _netlistHG.getNumNodes();

    abkfatal(cHG.getNumNodes()-firstConstNode == _allBlocks.size()*4,
	"mismatch in #nodes between constraining HG and #Blocks");

    n = firstConstNode;
    for(unsigned b = 0; b < _allBlocks.size(); b++)
    {
	const BBox& blockBBox = _allBlocks[b]->getBBox();
	nodesToFix.push_back(n);
	pl[n++] = Point(blockBBox.xMin, blockBBox.yMin); 
	nodesToFix.push_back(n);
	pl[n++] = Point(blockBBox.xMin, blockBBox.yMax); 
	nodesToFix.push_back(n);
	pl[n++] = Point(blockBBox.xMax, blockBBox.yMax); 
	nodesToFix.push_back(n);
	pl[n++] = Point(blockBBox.xMax, blockBBox.yMin); 
    }

    Mapping fixedMap(cHG.getNumTerminals() + _allBlocks.size()*4,
		cHG.getNumNodes(), nodesToFix);

    Placement fixedLocs(fixedMap, pl);
    FixedGroupConstraint fxConst(pl, fixedMap, fixedLocs);
    Constraints bplConst;
    bplConst.add(fxConst);


	//call BoxPlacer on the constrained placment problem
     _boxPl = new BBoxPlacer(cHG, bplConst, pl, params);
}

BoxPlacementForCapo::~BoxPlacementForCapo()
{ 
    if(_boxPl) delete _boxPl; _boxPl = NULL;
}

void BoxPlacementForCapo::setPlacement(Placement& plToSet)
{
    if(_boxPl)
    {
        for(unsigned n = 0; n < plToSet.size(); n++)
            plToSet[n] = (*_boxPl)[n];
    }
}

bool BoxPlacementForCapo::placementIsValid()
{
	//checks that each cell that is inside a box actually IS
	//inside the block

    bool isValid = true;

    for(unsigned b = 0; b < _allBlocks.size(); b++)
    {
	const BBox& blockBBox = _allBlocks[b]->getBBox();
	vector<unsigned>::const_iterator cId;
	for(cId  = _allBlocks[b]->cellIdsBegin();
	    cId != _allBlocks[b]->cellIdsEnd(); cId++)
   	{
	    if(_boxPl && !blockBBox.contains((*_boxPl)[*cId])) 
	    {
		isValid = false;
	        cout<<"Node Placed Outside Block: "<<blockBBox<<" - "<<
			(*_boxPl)[*cId]<<endl;
	    }
	}
    }

    return isValid;
}

